﻿using System.Numerics;
using Nebulix;
using Nebulix.Rendering;
using Silk.NET.OpenGL;

namespace Engine_silk.NET;

public class IcoSphere
{
    private readonly Mesh mesh = new();
    private readonly int resolution;
    private readonly IcoSphereGenerator generator = new();
    
    public IcoSphere(int resolution)
    {
        if (resolution < 2)
            throw new ArgumentOutOfRangeException(nameof(resolution), resolution,
                "Resolution must be greater than 1");
        this.resolution = resolution;
    }

    public void CreateSphere()
    {
        generator.Generate(resolution);
        
        mesh.Clear();
        mesh.Vertices = generator.Vertices;
        mesh.Indices = Array.ConvertAll(generator.Triangles, x => (uint)x);
        mesh.CalculateNormals();
    }
    
    public void RenderSphere(GL gl) // Will not be needed  
    {
        mesh.Render(gl);
    }
}

// Thank you Sebastian Lague
class IcoSphereGenerator 
{

    // Output:
    public Vector3[] Vertices => vertices?.Items ?? Array.Empty<Vector3>();
    public int[] Triangles => triangles?.Items ?? Array.Empty<int>();

    // Internal:
    FixedSizeList<Vector3>? vertices;
    FixedSizeList<int>? triangles;
    int numDivisions;
    int numVertsPerFace;

    // Indices of the vertex pairs that make up each of the initial 12 edges
    static readonly int[] VertexPairs = { 0, 1, 0, 2, 0, 3, 0, 4, 1, 2, 2, 3, 3, 4, 4, 1, 5, 1, 5, 2, 5, 3, 5, 4 };
    // Indices of the edge triplets that make up the initial 8 faces
    static readonly int[] EdgeTriplets = { 0, 1, 4, 1, 2, 5, 2, 3, 6, 3, 0, 7, 8, 9, 4, 9, 10, 5, 10, 11, 6, 11, 8, 7 };
    // The six initial vertices                up             left           back            right           forward        down
    static readonly Vector3[] BaseVertices = { Vector3.UnitY, -Vector3.UnitX, -Vector3.UnitZ, Vector3.UnitX, Vector3.UnitZ, -Vector3.UnitY };

    public void Generate (int resolution) {
        numDivisions = Math.Max (0, resolution);
        numVertsPerFace = ((numDivisions + 3) * (numDivisions + 3) - (numDivisions + 3)) / 2;
        int numVerts = numVertsPerFace * 8 - (numDivisions + 2) * 12 + 6;
        int numTrisPerFace = (numDivisions + 1) * (numDivisions + 1);

        vertices = new FixedSizeList<Vector3> (numVerts);
        triangles = new FixedSizeList<int> (numTrisPerFace * 8 * 3);

        vertices.AddRange (BaseVertices);

        // Create 12 edges, with n vertices added along them (n = numDivisions)
        Edge[] edges = new Edge[12];
        for (int i = 0; i < VertexPairs.Length; i += 2) {
            Vector3 startVertex = vertices.Items[VertexPairs[i]];
            Vector3 endVertex = vertices.Items[VertexPairs[i + 1]];

            int[] edgeVertexIndices = new int[numDivisions + 2];
            edgeVertexIndices[0] = VertexPairs[i];

            // Add vertices along edge
            for (int divisionIndex = 0; divisionIndex < numDivisions; divisionIndex++) {
                float t = (divisionIndex + 1f) / (numDivisions + 1f);
                edgeVertexIndices[divisionIndex + 1] = vertices.NextIndex;
                vertices.Add (startVertex.Slerp(endVertex, t));
            }
            edgeVertexIndices[numDivisions + 1] = VertexPairs[i + 1];
            int edgeIndex = i / 2;
            edges[edgeIndex] = new Edge (edgeVertexIndices);
        }

        // Create faces
        for (int i = 0; i < EdgeTriplets.Length; i += 3) {
            int faceIndex = i / 3;
            bool reverse = faceIndex >= 4;
            CreateFace (edges[EdgeTriplets[i]], edges[EdgeTriplets[i + 1]], edges[EdgeTriplets[i + 2]], reverse);
        }
    }

    void CreateFace (Edge sideA, Edge sideB, Edge bottom, bool reverse)
    {
        if (vertices is null) throw new ArgumentException("Vertices cannot be null", nameof(vertices));
        if (triangles is null) throw new ArgumentException("triangles cannot be null", nameof(triangles));
        
        int numPointsInEdge = sideA.VertexIndices.Length;
        var vertexMap = new FixedSizeList<int> (numVertsPerFace);
        vertexMap.Add (sideA.VertexIndices[0]); // top of triangle

        for (int i = 1; i < numPointsInEdge - 1; i++) {
            // Side A vertex
            vertexMap.Add (sideA.VertexIndices[i]);

            // Add vertices between sideA and sideB
            Vector3 sideAVertex = vertices.Items[sideA.VertexIndices[i]];
            Vector3 sideBVertex = vertices.Items[sideB.VertexIndices[i]];
            int numInnerPoints = i - 1;
            for (int j = 0; j < numInnerPoints; j++) {
                float t = (j + 1f) / (numInnerPoints + 1f);
                vertexMap.Add (vertices.NextIndex);
                vertices.Add (sideAVertex.Slerp(sideBVertex, t));
            }

            // Side B vertex
            vertexMap.Add (sideB.VertexIndices[i]);
        }

        // Add bottom edge vertices
        for (int i = 0; i < numPointsInEdge; i++) {
            vertexMap.Add (bottom.VertexIndices[i]);
        }

        // Triangulate
        int numRows = numDivisions + 1;
        for (int row = 0; row < numRows; row++) {
            // vertices down left edge follow quadratic sequence: 0, 1, 3, 6, 10, 15...
            // the nth term can be calculated with: (n^2 - n)/2
            int topVertex = ((row + 1) * (row + 1) - row - 1) / 2;
            int bottomVertex = ((row + 2) * (row + 2) - row - 2) / 2;

            int numTrianglesInRow = 1 + 2 * row;
            for (int column = 0; column < numTrianglesInRow; column++) {
                int v0, v1, v2;

                if (column % 2 == 0) {
                    v0 = topVertex;
                    v1 = bottomVertex + 1;
                    v2 = bottomVertex;
                    topVertex++;
                    bottomVertex++;
                } else {
                    v0 = topVertex;
                    v1 = bottomVertex;
                    v2 = topVertex - 1;
                }

                triangles.Add (vertexMap.Items[v0]);
                triangles.Add (vertexMap.Items[(reverse) ? v2 : v1]);
                triangles.Add (vertexMap.Items[(reverse) ? v1 : v2]);
            }
        }

    }

    // Convenience classes:

    private record Edge(int[] VertexIndices);

    private class FixedSizeList<T>(int size)
    {
        public readonly T[] Items = new T[size];
        public int NextIndex;

        public void Add (T item) {
            Items[NextIndex] = item;
            NextIndex++;
        }

        public void AddRange (IEnumerable<T> items) {
            foreach (var item in items) {
                Add (item);
            }
        }
    }

}