﻿using System.Numerics;
using Nebulix;
using Nebulix.Rendering;
using Silk.NET.OpenGL;

namespace Engine_silk.NET
{
	// also maybe make an ISphere and call this class "CubeSphere"?
	public class Sphere
	{
		private readonly Mesh mesh = new();
		private readonly uint resolution;
		
		public Sphere(uint resolution)
		{
			if (resolution < 2)
				throw new ArgumentOutOfRangeException(nameof(resolution), resolution,
					"Resolution must be greater than 1");
			this.resolution = resolution;
		}

		public void CreateSphere()
		{
			// TODO: merge the individual meshes to one mesh
			Vector3[] directions =
			[
				Vector3.UnitZ, -Vector3.UnitZ, 
				Vector3.UnitY, -Vector3.UnitY,
				Vector3.UnitX, -Vector3.UnitX
			];

			List<Vector3> vertices = new();
			List<uint> indices = new();
			for (int i = 0; i < directions.Length; i++)
			{
				Mesh m = new();
				Face f = new(directions[i], m, resolution);
				f.GenerateMesh();

				vertices.AddRange(m.Vertices);
				for (int j = 0; j < m.Indices.Length; j++)
				{
					indices.Add(m.Indices[j] + (uint)(m.Vertices.Length * i));
				}
			}
			// TODO get rid of overlapping vertices maybe
			mesh.Clear();
			mesh.Vertices = vertices.ToArray();
			mesh.Indices = indices.ToArray();
			mesh.CalculateNormals();
		}
		
		/// <summary>
		/// Renders the mesh. No Shader is specified and <see cref="CreateSphere"/> needs to be called beforehand 
		/// </summary>
		public void RenderSphere(GL gl) // Will not be needed  
		{
			mesh.Render(gl);
		}
	}

	internal record struct Face
	{
		public Mesh Mesh => _mesh;

		private readonly Mesh _mesh;
		private readonly Vector3 _localX;
		private readonly Vector3 _localY;

		private readonly Vector3 _localUp;
		private readonly uint _resolution;
		
		public Face(Vector3 localUp, Mesh mesh, uint resolution)
		{
			_mesh = mesh;
			_localX = new(localUp.Y, localUp.Z, localUp.X);
			_localY = Vector3.Cross(localUp, _localX);
			_localUp = localUp;
			_resolution = resolution;
		}

		internal void GenerateMesh()
		{
			var vertices = new Vector3[_resolution * _resolution];
			// _resolution - 1 because the vertices index starts at 0 
			// * 6 because each triangle needs 3 points and each small quad has 2 triangles 3*2 = 6
			var indices = new uint[(_resolution - 1) * (_resolution - 1) * 6];
			int triangleIndex = 0;

			uint i;
			for (uint y = 0; y < _resolution; y++)
			{
				for (uint x = 0; x < _resolution; x++)
				{
					i = x + y * _resolution;
					Vector2 percent = new Vector2(x, y) / (_resolution - 1);
					// place vertex on correct position of the plane to easily calculate indices
					Vector3 vertexPosition = _localUp + (percent.X - 0.5f) * 2 * _localX + (percent.Y - 0.5f) * 2 * _localY;
					vertices[i] = PointOnCubeToPointOnSphere(vertexPosition);

					if (x != _resolution - 1 && y != _resolution - 1) // we didn't reach the bottom right point yet
					{
						indices[triangleIndex] = i;
						indices[triangleIndex + 1] = i + _resolution + 1;
						indices[triangleIndex + 2] = i + _resolution;
						
						indices[triangleIndex + 3] = i;
						indices[triangleIndex + 4] = i + 1;
						indices[triangleIndex + 5] = i + _resolution + 1;

						triangleIndex += 6; // doing it this way is faster than doing indices[triangleIndex++] (takes about double the time)
					}
				}
			}

			_mesh.Clear();
			_mesh.Vertices = vertices;
			_mesh.Indices = indices;
			_mesh.CalculateNormals();
		}

		// Smooth mapping so that the points are not clumped on the former corners of the cube
		// http://mathproofs.blogspot.com/2005/07/mapping-cube-to-sphere.html
		private Vector3 PointOnCubeToPointOnSphere(Vector3 point)
		{
			float x2 = point.X * point.X;
			float y2 = point.Y * point.Y;
			float z2 = point.Z * point.Z;
			float newX = point.X * MathF.Sqrt(1 - y2 / 2 - z2 / 2 + (y2 * z2) / 3);
			float newY = point.Y * MathF.Sqrt(1 - z2 / 2 - x2 / 2 + (z2 * x2) / 3);
			float newZ = point.Z * MathF.Sqrt(1 - x2 / 2 - y2 / 2 + (x2 * y2) / 3);
			return new Vector3(newX, newY, newZ);
		}
	}
}
